#! /usr/bin/env node

const shell = require("shelljs");
const replaceFN = require("replace-in-file");
const { execSync } = require("child_process");
const fs = require("fs");
const path = require("path");
var isValidGlob = require("is-valid-glob");
const yargs = require("yargs/yargs");

const { hideBin } = yargs;
const argv = yargs(hideBin(process.argv)).argv;

if (argv["staged-only"] && !shell.which("git")) {
  shell.echo("Sorry, --staged-only feature requires git");
  shell.exit(1);
}

const specificReplaces = [
  {
    from: /from "@mui\/system/,
    to: `from "@mui/material`,
  },
  {
    from: /import (\w+)Props from "@mui\/material\/\w+Props"/,
    to: `import {$1Props} from "@mui/material/$1"`,
  },
  {
    from: `import styled from "@mui/material/styled"`,
    to: `import { styled } from "@mui/material/styles"`,
  },
];

function generateRegex(count) {
  return {
    find: `import \\{[\\s\\n]*${new Array(count)
      .fill("(\\w+),?")
      .join("[\\s\\n]+")}[\\s\\n]*\\}[\\s\\n]+from\\s+['"]@mui/([a-z-]+)['"]`,
    replace: new Array(count)
      .fill("")
      .map((n, i) => `import $${i + 1} from "@mui/$${count + 1}/$${i + 1}"`)
      .join("\n"),
  };
}

function generateFontAwesomeRegex(count) {
  return {
    find: `import \\{[\\s\\n]*${new Array(count)
      .fill("(\\w+),?")
      .join(
        "[\\s\\n]+"
      )}[\\s\\n]*\\}[\\s\\n]+from\\s+['"]@fortawesome/free-([a-z-]+)-svg-icons['"]`,
    replace: new Array(count)
      .fill("")
      .map(
        (n, i) =>
          `import { $${i + 1} } from "@fortawesome/free-$${
            count + 1
          }-svg-icons/$${i + 1}"`
      )
      .join("\n"),
  };
}

/**
 *
 * @param {string[]} pathsToWorkWith
 */
function run(pathsToWorkWith) {
  const filesChanged = new Set();

  console.log("🌳 Tree Shaking MUI imports");
  const muiRegex = new Array(25).fill(null).map((_, i) => generateRegex(i));
  replaceFN.sync({
    ignore: ["node_modules/*"],
    files: pathsToWorkWith,
    from: muiRegex.map(({ find }) => new RegExp(find, "g")),
    to: muiRegex.map(({ replace, find }, index) => (match, ...args) => {
      if (match) {
        const filePath = args[args.length - 1].replace(process.cwd() + "/", "");
        console.log(`🪓 Cutting in "${filePath}"`);
        filesChanged.add(filePath);
        return match.replace(new RegExp(find, "g"), replace);
      }
      return match;
    }),
  });

  console.log("🌳 Tree Shaking Fontawesome imports");
  const fontAwesomeRegex = new Array(25)
    .fill(null)
    .map((_, i) => generateFontAwesomeRegex(i));
  replaceFN.sync({
    ignore: ["node_modules/*"],
    files: pathsToWorkWith,
    from: fontAwesomeRegex.map(({ find }) => new RegExp(find, "g")),
    to: fontAwesomeRegex.map(({ replace, find }, index) => (match, ...args) => {
      console.log(match);
      if (match) {
        const filePath = args[args.length - 1].replace(process.cwd() + "/", "");
        console.log(`🪓 Cutting in "${filePath}"`);
        filesChanged.add(filePath);
        return match.replace(new RegExp(find, "g"), replace);
      }
      return match;
    }),
  });

  console.log("⚒️ Running Cleanups");
  replaceFN.sync({
    ignore: ["node_modules/*"],
    files: pathsToWorkWith,
    from: specificReplaces.map(({ from }) => new RegExp(from, "g")),
    to: specificReplaces.map(({ from, to }, index) => (match, ...args) => {
      if (match) {
        const filePath = args[args.length - 1].replace(process.cwd() + "/", "");
        console.log(`🪓 Cutting in "${filePath}"`);
        filesChanged.add(filePath);
        return match.replace(new RegExp(from, "g"), to);
      }
      return match;
    }),
  });

  console.log(`📳 ${filesChanged.size} files has been shaken`);
  console.log("✅ All Trees has been pruned ");
  console.log("");
  return Array.from(filesChanged.values());
}

function getChangedFiles(extension) {
  const extensionFilter = extension ? `-- '***.${extension}'` : "";
  const command = `git diff HEAD --name-only ${extensionFilter}`;
  const diffOutput = execSync(command).toString();
  return diffOutput.toString().split("\n").filter(Boolean);
}

const init = () => {
  const files = [];

  if (argv["staged-only"]) {
    const jsFiles = getChangedFiles("js");
    const tsFiles = getChangedFiles("ts");
    const tsxFiles = getChangedFiles("tsx");
    files.push(...jsFiles, ...tsFiles, ...tsxFiles);
  } else if (argv["_"]?.length) {
    const specificPaths = argv["_"];
    files.push(...specificPaths);
  }

  const allFiles = files
    .map((p) => {
      // making sure all paths are absolute
      if (path.isAbsolute(p)) {
        return p;
      }
      return path.resolve(process.cwd() + "/" + p);
    })
    .flatMap((p) => {
      // making sure that all paths are direct files or globs, and not a directory

      // if it is a glob pattern, then return as is;
      if (!isValidGlob(p)) {
        return [p];
      }

      // if path doesn't exist, then ignore
      if (!fs.existsSync(p)) {
        return [];
      }

      // if it is a directory
      const pathStat = fs.statSync(p);
      if (pathStat.isDirectory()) {
        return ["**/*.js", "**/*.ts", "**/*.tsx"].map((ext) => p + "/" + ext);
      }

      // if it is a specific file
      return [p];
    });

  const changedFiles = run(allFiles);
  if (argv["staged-only"]) {
    execSync(`git add ${changedFiles.join(" ")}`);
  }
};

init();
